!find_rotation
!______________________________________________________________________
! rotate horizontal components of a receiver gather after NMO with vel=water and sum along offset axis
!______________________________________________________________________
! USAGE:
! find_rotation2.x < in.H > out.H pars
!______________________________________________________________________
program find_rotation2
implicit none
!______________________________________________________________________
! allocatable arrays
!______________________________________________________________________
real,		allocatable, dimension(:,:) :: headers				! header block, expect all keys to be floats
real,		allocatable, dimension(:,:) :: data					! input seismic data
real,		allocatable, dimension(:,:) :: vx_data, vy_data		! X and Y data components
real,		allocatable, dimension(:,:,:)   :: vx_rot, vy_rot		! X and Y data components rotated at all angles
real,		allocatable, dimension(:)   :: sx,sy,gy,gx			! group coordinates
real,		allocatable, dimension(:)   :: nodeid				! receiver box number
real,		allocatable, dimension(:)   :: tracf				! component number
real,		allocatable, dimension(:)   :: offx					! X offset 
integer,	allocatable, dimension(:)   :: drn					! data record number
real,		allocatable, dimension(:)   :: sg_azi_vx, sg_azi_vy		! source-geophone azimuth angle
!______________________________________________________________________
! LOCATION OF VARIOUS KEYS IN THE HEADER STRUCTURES
!______________________________________________________________________
integer :: sx_key, sy_key, gy_key, gx_key, node_key, tracf_key, offx_key, drn_key
!______________________________________________________________________
! simple variables
!______________________________________________________________________
integer :: nt, nsrc, nang	! number of input samples, and number of output sources and rotation angles
real    :: ot, osrc, oang
real    :: dt, dsrc, dang
real	:: minoff, maxoff	! minimum and maximum source-receiver offsets for which we want to look at rotation angles.
integer :: nkeys			! number of keys
integer :: ntrace			! total number of input traces
integer :: maxsize			! memory available for storage in words.
integer :: maxtr_per_block	! maximum number of traces in a block of data.
integer :: ntr_per_block	! number of traces in current block of data.
integer :: nblocks			! number of blocks of needed to read all the traces.
integer :: jblock			! block index
integer :: fwind			! first trace of the data window to be read.
integer :: ierr				! return value checker
logical :: have_drn			! WHETHER OR NOT DRN EXISTS
real	:: node_num			! id of node we want to rotate
logical :: verb
real	:: ang
real	:: eps=1e-4
integer	:: iang, isrc, isrc_vx, isrc_vy
real, parameter :: VX=1, VY=2, deg2rad=4.*atan(1.)/180.
!______________________________________________________________________
! control variables
!______________________________________________________________________
integer :: allocation_status ! return code for Fortran90 allocate command
integer :: stderr=0 ! standard error unit
!______________________________________________________________________
! DUMMY VARIALBES
!______________________________________________________________________
real :: o,dd
integer :: itrc ! LOOP VARIABLE
character(len=1025) :: label

!EXTERNAL SHOULD BE USED FOR ALL C CALLS, SOME F90 COMPILERS GET MAD WITHOUT IT
integer,external :: getch				! SEPlib get param from parameter file
integer,external :: hetch				! SEPlib get param from history file
integer,external :: fetch				! SEPlib get from either parameter or history file
integer,external :: putch				! SEPlib put to history file
integer,external :: sep_get_number_keys	! SEPlib number of keys
integer,external :: sep_get_key_index	! SEPlib index of given header key
integer,external :: sseek_block			! SEPlib SEEK FILE ROUTINE
integer,external :: sreed				! SEPlib READ ROUTINE
integer,external :: srite				! SEPlib WRITE ROUTINE
integer,external :: sep_get_header_axis_par		! SEPlib HEADER AXIS ROUTINE
integer,external :: sep_get_number_header_axes	! SEPlib HEADER AXIS ROUTINE

!______________________________________________________________________
! self documentation
!______________________________________________________________________
call sep_add_doc_line("NAME")
call sep_add_doc_line("    find_rotation - rotate the two horizontal components of a receiver gather around the vertical axis within a specified range of angles.")
call sep_add_doc_line("")
call sep_add_doc_line("SYNOPSIS")
call sep_add_doc_line("    find_rotation.x < in.H > out.H pars")
call sep_add_doc_line("")
call sep_add_doc_line("INPUT PARAMETERS")
call sep_add_doc_line("    verb - integer")
call sep_add_doc_line("               [0] Whether (1) or not(0) to be verbose")
call sep_add_doc_line("")
call sep_add_doc_line("    maxsize - integer")
call sep_add_doc_line("               [10] Number of megabytes of traces to hold")
call sep_add_doc_line("")
call sep_add_doc_line("    nodeid - integer")
call sep_add_doc_line("               [0] -  ocean-bottom node (receiver) physical box number")
call sep_add_doc_line("")
call sep_add_doc_line("    nsrc - integer")
call sep_add_doc_line("               [0]  number of shots in receiver gather")
call sep_add_doc_line("")
call sep_add_doc_line("    dsrc - float")
call sep_add_doc_line("               [0.]  shot spacing")
call sep_add_doc_line("")
call sep_add_doc_line("    osrc - float")
call sep_add_doc_line("               [0.]  shot origin")
call sep_add_doc_line("")
call sep_add_doc_line("    nang - integer")
call sep_add_doc_line("               [0]  number of angles of rotation")
call sep_add_doc_line("")
call sep_add_doc_line("    dang - float")
call sep_add_doc_line("               [0.]  angle spacing")
call sep_add_doc_line("")
call sep_add_doc_line("    oang - float")
call sep_add_doc_line("               [0.]  angle origin")
call sep_add_doc_line("")
call sep_add_doc_line("    minoff - float")
call sep_add_doc_line("               [0.]  minimum offset to sum along gather from")
call sep_add_doc_line("")
call sep_add_doc_line("    maxoff - float")
call sep_add_doc_line("               [0.]  maximum offset to sum along gather until")
call sep_add_doc_line("")
call sep_add_doc_line("DESCRIPTION")
call sep_add_doc_line("     The expected input is a receiver gather that has had NMO applied to it with water velocity.")
call sep_add_doc_line("     The program rotates the horizontal components of the specified node (receiver), within the specified angle range.")
call sep_add_doc_line("     It then sums absoulte values along the offset axis within the specified offset range.")
call sep_add_doc_line("     The output is a file of (nangles)x2, that shows the sum of the traces after rotation by each angle for the two horizontal components.")
call sep_add_doc_line("")
call sep_add_doc_line("COMMENTS")
call sep_add_doc_line("     The program assumes the sources are sorted by their X position. Also, the following keywords are assumed:")
call sep_add_doc_line("     Node id key = Nodeid")
call sep_add_doc_line("     receiver x position key = Gxfb")
call sep_add_doc_line("     receiver y position key = Gyfb")
call sep_add_doc_line("     source x position key = Sx")
call sep_add_doc_line("     source y position key = Sy")
call sep_add_doc_line("     Component key = Tracf (1=Vx, 2=Vy)")
call sep_add_doc_line("     Offset key = Offsetx")
call sep_add_doc_line("")

!_______________________________________________________________
! initialize the input seismic data file.
! do i need to specify any arguments?
!_______________________________________________________________
!TO INITIALIZE SEP IO/ PARA HANDLING
call initpar()
!SO SELF DOC WORKS
call doc(SOURCE)
!TO INITIALiZE SEP3D IO
call init_3d()
!_______________________________________________________________
! read in important parameters from the history or parameter file
!_______________________________________________________________
if (hetch ('n1',       'd', nt)       <= 0) call seperr('cant read n1 from history file!')
if (hetch ('o1',       'f', ot)       <= 0) call seperr('cant read o1 from history file!')
if (hetch ('d1',       'f', dt)       <= 0) call seperr('cant read d1 from history file!')
if (getch ('maxsize',   'd', maxsize )  <= 0) maxsize=10
if (getch ('nodeid',   'f', node_num) <= 0) node_num=0. ! box number of node that we want to find the rotation for
if (getch ('nsrc',     'd', nsrc)     <= 0) nsrc=0
if (getch ('dsrc',     'f', dsrc)     <= 0) dsrc=0.
if (getch ('osrc',     'f', osrc)     <= 0) osrc=0.
if (getch ('nang',     'd', nang)     <= 0) nang=0
if (getch ('dang',     'f', dang)     <= 0) dang=0.
if (getch ('oang',     'f', oang)     <= 0) oang=0.
if (getch ('minoff',   'f', minoff)   <= 0) minoff=0.
if (getch ('maxoff',   'f', maxoff)   <= 0) maxoff=0.
if (getch ('verb',     'd', verb)     <= 0) verb=.false.

if (verb) then
	write (stderr,*) 'nodeid=', node_num
endif

!THE NUMBER OF TRACES IS IN THE HEADER FILE
!IT DOESN'T HAVE TO BE THE SAME IS IN THE HISTORY FILE
!THEREFORE
if(sep_get_number_header_axes("in",ierr) .ne. 0) call seperr("trouble obtaining the number of header axes")

if(ierr .ne. 2) call seperr("this program expects a 2-D header file")

! read from axis 2: number of traces
if(sep_get_header_axis_par("in", 2, ntrace, o, dd, label) .ne. 0) call seperr("trouble obtaining the number of headers")

!_______________________________________________________________
! read in important parameters from the hff file
!_______________________________________________________________
if (sep_get_number_keys("in",nkeys) .ne. 0) call seperr("trouble obtaining number of keys from hff")

if (sep_get_key_index("in", "Nodeid", node_key) .ne. 0) call seperr("trouble obtaining Nodeid key number")
if (sep_get_key_index("in", "Gxfb", gx_key) .ne. 0) call seperr("trouble obtaining Gxfb key number")
if (sep_get_key_index("in", "Gyfb", gy_key) .ne. 0) call seperr("trouble obtaining Gyfb key number")
if (sep_get_key_index("in", "Sx", sx_key) .ne. 0) call seperr("trouble obtaining Sx key number")
if (sep_get_key_index("in", "Sy", sy_key) .ne. 0) call seperr("trouble obtaining Sy key number")
if (sep_get_key_index("in", "Tracf", tracf_key) .ne. 0) call seperr("trouble obtaining Tracf key number")
if (sep_get_key_index("in", "Offsetx", offx_key) .ne. 0) call seperr("trouble obtaining Tracf key number")


ierr = sep_get_key_index("in","data_record_number",drn_key)

if (ierr .ne. 0) then !DRN DOESN'T EXIST;

	if (getch ("same_record_number","d",ierr) == 1) then !IF SAME RECORD NUMBER IN HISTORY
		if(ierr==0) call seperr("no data_record_number but traces out of order with data")
	end if
	have_drn=.false.
else
	have_drn=.true.
end if

!______________________________________________________________________
! determine how many traces can fit into maxsize words of memory.
! determine how many blocks will need to be read in to read through
! all the input data.
!______________________________________________________________________
maxtr_per_block=(maxsize*1024*1024-1)/(ntrace*nt)+1
nblocks=(ntrace-1)/maxtr_per_block+1

!______________________________________________________________________
! ALLOCATE arrays
! note that output arrays are of length nt (the same as the input).
! arrays used in Fourier transform are of length nt_fft >= nt.
!______________________________________________________________________
allocate (data (0:nt-1,maxtr_per_block), headers (nkeys,maxtr_per_block), &
	gy (maxtr_per_block), gx (maxtr_per_block), nodeid (maxtr_per_block), &
	sy (maxtr_per_block), sx (maxtr_per_block), tracf (maxtr_per_block), &
	offx (maxtr_per_block), drn (maxtr_per_block), &
	STAT=allocation_status)

if(allocation_status /= 0) THEN
	write (stderr,*) 'space for data,headers,gy,gx,nodeid,sy,sx,tracf could not be allocated!'
	write (stderr,*) 'program aborted!'
	stop 9666
endif

allocate (vx_data (0:nt-1,nsrc), vy_data (0:nt-1,nsrc), &
	sg_azi_vx (nsrc), sg_azi_vy (nsrc), &
	vx_rot (nt, nsrc, nang), vy_rot (nt, nsrc, nang), &
	STAT=allocation_status)

if(allocation_status /= 0) THEN
	write (stderr,*) 'space for vx and vy data could not be allocated!'
	write (stderr,*) 'program aborted!'
	stop 9667
endif

vx_rot = 0.
vy_rot = 0.
!_______________________________________________________________________
! write out the history files defining the size of the output.
!_______________________________________________________________________
!LETS DEAL WITH THE OUTPUT
!WE ARE GOING TO REORDER THE TRACES, SO LETS TELL THE PROGRAM THAT
!THE TRACES ARE IN THE SAME ORDER

!if(0 .ne. putch ("same_record_number","d",1)) call seperr("trouble putching same_record_number")


if(0 .ne. putch ("n1", "d",  nt)) call seperr("trouble putching nt")
if(0 .ne. putch ("o1", "f",  ot)) call seperr("trouble putching ot")
if(0 .ne. putch ("d1", "f",  dt)) call seperr("trouble putching dt")
if(0 .ne. putch ("n2", "d",  nsrc)) call seperr("trouble putching nsrc")
if(0 .ne. putch ("o2", "f",  osrc)) call seperr("trouble putching osrc")
if(0 .ne. putch ("d2", "f",  dsrc)) call seperr("trouble putching dsrc")
if(0 .ne. putch ("n3", "d",  nang)) call seperr("trouble putching nang")
if(0 .ne. putch ("o3", "f",  oang)) call seperr("trouble putching oang")
if(0 .ne. putch ("d3", "f",  dang)) call seperr("trouble putching dang")
if(0 .ne. putch ("n4", "d",  2))    call seperr("trouble putching ncomps")
if(0 .ne. putch ("o4", "f",  0.))   call seperr("trouble putching ocomps")
if(0 .ne. putch ("d4", "f",  1.))   call seperr("trouble putching dcomps")
if(0 .ne. putch ("hff","d", -1))    call seperr("trouble putching hff=-1")
if(0 .ne. putch ("gff","d", -1))    call seperr("trouble putching gff=-1")

!_____________________________________________________________
! close sep data files
!_____________________________________________________________
call hclose()

isrc_vx = 1
isrc_vy = 1
!___________________________________________________________________
! loop over the blocks of data.
!___________________________________________________________________
loop_over_blocks: do jblock=1,nblocks

	!___________________________________________________________________
	! calculate the number of traces in this block.
	!___________________________________________________________________
	ntr_per_block=maxtr_per_block
	if (jblock == nblocks) ntr_per_block = ntrace-(nblocks-1)*maxtr_per_block
	!___________________________________________________________________
	! grab the next ntr_per_block trace headers from the trace header file.
	! 'in' = standard in
	! headers = a fortran90 structure
	! fwind = the sequential trace position in the trace header file.
	!___________________________________________________________________
	fwind=(jblock-1)*maxtr_per_block

	!READ IN THE HEADERS IN THIS BLOCK
	call sep_get_val_headers("in", fwind+1, ntr_per_block, headers)
	!_______________________________________________________________________
	! echo out how many headers are out there on this file.
	!_______________________________________________________________________
	!___________________________________________________________________
	! grab a subset of desired headers using the header key words.
	!___________________________________________________________________
	nodeid	= headers (node_key,:)
	gy		= headers (gy_key,:)
	gx		= headers (gx_key,:)
	sy		= headers (sy_key,:)
	sx		= headers (sx_key,:)
	tracf	= headers (tracf_key,:)
	offx	= headers (offx_key,:)

	!IF WE HAVE A DATA RECORD NUMBER WE NEED TO GRAB THE
	!CORRESPONDING TRACE NUMBERS. THE HEADER BLOCK IS
	!REAL, DRN ARE INTS. WE THEREFORE NEED TO CONVERT

	if(have_drn) then
		drn(1:ntr_per_block) = transfer (headers (drn_key, :ntr_per_block), 0, ntr_per_block)
	
		do itrc=1,ntr_per_block
			!___________________________________________________________________
			! now let's try to read ntr_per_block traces of seismic data.
			!___________________________________________________________________
			!SEEK TO THE BEGINING OF THE GIVEN TRACE
			if (sseek_block ("in", drn(itrc)-1, 4*nt, 0) < 0)  call seperr("trouble seeking trace")
			if (sreed ("in", data (:,itrc),   4*nt) .ne. 4*nt) call seperr("trouble reading trace ")
			drn(itrc)=itrc+fwind !RESET DRN TO ITS NEW POSITION
		end do
	else !TRACES ARE IN ORDER SO JUST READ THEN
	!___________________________________________________________________
	! now let's try to read ntr_per_block traces of seismic data.
	!___________________________________________________________________
		if (sreed("in", data, 4*nt*ntr_per_block) .ne. 4*nt*ntr_per_block) call seperr("trouble reading trace block")
	end if

	!___________________________________________________________________
	! if this node is the one specified on the command line, extract the
	! Vx and Vy components and save to dedicated arrays.
	!___________________________________________________________________

	do itrc = 1, ntr_per_block
		if (nodeid (itrc) .lt. node_num+eps .and. nodeid (itrc) .gt. node_num-eps .and. abs(offx(itrc)) .gt. minoff .and. abs(offx(itrc)) .lt. maxoff) then
			if (verb) write (stderr,*) 'node=', nodeid (itrc),'sx=', sx (itrc), 'sy=', sy (itrc), 'gx=', gx (itrc), 'gy=', gy (itrc), 'offx=', offx(itrc), 'tracf=', tracf (itrc)

			if ((tracf (itrc) .lt. VX+eps .and. tracf (itrc) .gt. VX-eps .or. tracf (itrc) .lt. VY+eps .and. tracf (itrc) .gt. VY-eps) .and. isrc_vx .le. nsrc .and. isrc_vy .le. nsrc) then
				if (tracf (itrc) .lt. VX+eps) then
					vx_data (:,isrc_vx) = data (:,itrc)
					sg_azi_vx (isrc_vx) = atan2 ((sy(itrc) - gy(itrc)), (sx(itrc) - gx(itrc)))
					isrc_vx = isrc_vx + 1
				else
					vy_data (:,isrc_vy) = data (:,itrc)
					sg_azi_vy (isrc_vy) = atan2 ((sy(itrc) - gy(itrc)), (sx(itrc) - gx(itrc)))
					isrc_vy = isrc_vy + 1
				endif
			endif
			
		endif
	end do
	
end do loop_over_blocks

!_____________________________
! rotate X and Y components
!_____________________________

do iang = 1, nang
	ang = (oang + (iang-1)*dang) * deg2rad
	do isrc = 1, nsrc
		vx_rot (:, isrc, iang) = vx_data (:, isrc) * cos(ang) - vy_data (:, isrc) * sin(ang)
		vy_rot (:, isrc, iang) = vx_data (:, isrc) * sin(ang) + vy_data (:, isrc) * cos(ang)
	end do
end do


if (srite ("out", vx_rot, 4*nt*nsrc*nang) .ne. 4*nt*nsrc*nang) call seperr("trouble writing out data")
if (srite ("out", vy_rot, 4*nt*nsrc*nang) .ne. 4*nt*nsrc*nang) call seperr("trouble writing out data")

if (verb) write (stderr,*) 'find_rotation2: normal completion.'

end program find_rotation2

