!<
!read_geometry
!______________________________________________________________________
! read in some data and extract some headers using SEPlib3d
!______________________________________________________________________
! USAGE:
! read_geometry < in.H > out.H
!%end of self-documentation
!______________________________________________________________________
program fix_pol
!______________________________________________________________________
! access the sep module
!______________________________________________________________________
IMPLICIT NONE
!______________________________________________________________________
! allocatable arrays
!______________________________________________________________________
REAL,		ALLOCATABLE, DIMENSION(:,:) :: d		! input seismic data
REAL,		ALLOCATABLE, DIMENSION(:,:) :: headers	! HEADER BLOCK
REAL,		ALLOCATABLE, DIMENSION(:) :: gy			! group coordinates
REAL,		ALLOCATABLE, DIMENSION(:) :: nodeid		! id of ocean bottom node
REAL,		ALLOCATABLE, DIMENSION(:) :: tracf		! component index (1=Vx/X, 2=Vy/Y, 3=Vz/Z, 4=Hyd/P, according to ebcdic header)
integer,	ALLOCATABLE, DIMENSION(:) :: drn		! data record number
!______________________________________________________________________
! LOCATION OF VARIOUS KEYS IN THE HEADER STRUCTURES
!______________________________________________________________________
integer ::gy_key, tracf_key, node_key, drn_key
!______________________________________________________________________
! simple variables
!______________________________________________________________________
integer :: nt				! number of input samples
integer :: nkeys			! number of keys
integer :: ntrace			! total number of input traces
integer :: nh				! number of headers per trace
integer :: maxmem			! memory available for storage in words.
integer :: maxtr_per_block	! maximum number of traces in a block of data.
integer :: ntr_per_block	! number of traces in current block of data.
integer :: nblocks			! number of blocks of needed to read all the traces.
integer :: jblock			! block index
integer :: fwind			! first trace of the data window to be read.
integer :: ierr				! return value checker
logical :: have_drn			! WHETHER OR NOT DRN EXISTS
real	:: gyflip
real	:: eps = 1e-4
real, parameter	:: VX=1, VY=2
logical :: verb
!______________________________________________________________________
! control variables
!______________________________________________________________________
integer :: allocation_status ! return code for Fortran90 allocate command
integer :: stderr=0 ! standard error unit
!______________________________________________________________________
! DUMMY VARIALBES
!______________________________________________________________________
real :: o,dd
integer :: i, itrc ! LOOP VARIABLE
character(len=1025) :: label

!EXTERNAL SHOULD BE USED FOR ALL C CALLS, SOME F90 COMPILERS GET MAD WITHOUT IT
integer,external :: getch				! SEPlib get param from parameter file
integer,external :: hetch				! SEPlib get param from history file
integer,external :: fetch				! SEPlib get from either parameter or history file
integer,external :: putch				! SEPlib put to history file
integer,external :: sep_get_number_keys	! SEPlib number of keys
integer,external :: sep_get_key_index	! SEPlib index of given header key
integer,external :: sseek_block			! SEPlib SEEK FILE ROUTINE
integer,external :: sreed				! SEPlib READ ROUTINE
integer,external :: srite				! SEPlib WRITE ROUTINE
integer,external :: sep_get_header_axis_par		! SEPlib HEADER AXIS ROUTINE
integer,external :: sep_get_number_header_axes	! SEPlib HEADER AXIS ROUTINE
!_______________________________________________________________
! initialize the input seismic data file.
! do i need to specify any arguments?
!_______________________________________________________________
!TO INITIALIZE SEP IO/ PARA HANDLING
call initpar()
!SO SELF DOC WORKS
call doc(SOURCE)
!TO INITIALiZE SEP3D IO
call init_3d()
!_______________________________________________________________
! read in important parameters from the history or parameter file
!_______________________________________________________________
if (hetch('n1','d',nt ) <= 0 ) call seperr('cant read n1 from history file!')
if (getch('maxmem','d',maxmem ) <= 0 ) maxmem=1000
if (getch('gyflip','f',gyflip) <= 0 ) gyflip=0. ! y coordinate above which we will flip polarity of the data of the vx and vy components.
if (getch('verb','d',verb) <= 0 ) verb=.false.

!THE NUMBER OF TRACES IS IN THE HEADER FILE
!IT DOESN'T HAVE TO BE THE SAME IS IN THE HISTORY FILE
!THEREFORE
if(sep_get_number_header_axes("in",ierr) .ne. 0) call seperr("trouble obtaining the number of header axes")

if(ierr .ne. 2) call seperr("for right now only works with 2-D header file")

! read from axis 2: number of traces
if(sep_get_header_axis_par("in", 2, ntrace, o, dd, label) .ne. 0) call seperr("trouble obtaining the number of headers")
!_______________________________________________________________
! read in important parameters from the hff file
!_______________________________________________________________
if (sep_get_number_keys("in",nkeys) .ne. 0) call seperr("trouble obtaining number of keys from hff")

if (sep_get_key_index("in", "Gyfb", gy_key)     .ne. 0) call seperr("trouble obtaining Gyfb key number")
if (sep_get_key_index("in", "Tracf", tracf_key) .ne. 0) call seperr("trouble obtaining Tracf key number")
if (sep_get_key_index("in", "Nodeid", node_key) .ne. 0) call seperr("trouble obtaining Nodeid key number")

ierr = sep_get_key_index("in","data_record_number",drn_key)

if (ierr .ne. 0) then !DRN DOESN'T EXIST;

	if (getch ("same_record_number","d",ierr) == 1) then !IF SAME RECORD NUMBER IN HISTORY
		if(ierr==0) call seperr("no data_record_number but traces out of order with data")
	end if
	have_drn=.false.
else
	have_drn=.true.
end if

!______________________________________________________________________
! determine how many traces can fit into maxmem words of memory.
! determine how many blocks will need to be read in to read through
! all the input data.
!______________________________________________________________________
maxtr_per_block=(maxmem*1024.*1024.-1.)/(ntrace*nt)+1
nblocks=(ntrace-1)/maxtr_per_block+1

!______________________________________________________________________
! ALLOCATE arrays
! note that output arrays are of length nt (the same as the input).
! arrays used in Fourier transform are of length nt_fft >= nt.
!______________________________________________________________________
allocate (d(0:nt-1,maxtr_per_block),headers(nkeys,maxtr_per_block), &
	gy(maxtr_per_block), tracf(maxtr_per_block), nodeid (maxtr_per_block), drn(maxtr_per_block), &
	STAT=allocation_status)

if(allocation_status /= 0) THEN
	write (stderr,*) 'space for d,gy,tracf,nodeid could not be allocated!'
	write (stderr,*) 'program aborted!'
	stop 9666
endif
!_______________________________________________________________________
! write out the history files defining the size of the output.
!_______________________________________________________________________
!LETS DEAL WITH THE OUTPUT
!WE ARE GOING TO REORDER THE TRACES, SO LETS TELL THE PROGRAM THAT
!THE TRACES ARE IN THE SAME ORDER

if(0 .ne. putch ("same_record_number","d",1)) call seperr("trouble putching same_record_number")

!IN THIS PROGRAM THE OUTPUT IS THE SAME SIZE OF
!THE INPUT. THE HISTORY FILE BY DEFAULT IS THE SAME
!SIZE AS THE INPUT SO WE DON'T NEED TO DO ANY MANIPULATION
!
!THE HFF FILE DOES NEED TO BE CREATED. SAYING WE
!HAVE THE SAME KEYS. WE CAN JUST COPY THE HFF FILE
call sep_copy_hff("in","out")

!_____________________________________________________________
! close sep data files
!_____________________________________________________________
call hclose()
!___________________________________________________________________
! loop over the blocks of data.
!___________________________________________________________________
loop_over_blocks: do jblock=1,nblocks

	!___________________________________________________________________
	! calculate the number of traces in this block.
	!___________________________________________________________________
	ntr_per_block=maxtr_per_block
	if (jblock == nblocks) ntr_per_block = ntrace-(nblocks-1)*maxtr_per_block
	!___________________________________________________________________
	! grab the next ntr_per_block trace headers from the trace header file.
	! 'in' = standard in?
	! headers = a fortran90 structure
	! nh = the number of headers in the structure.
	! fwind = the sequential trace position in the trace header file.
	! or is the area where the next position on the file will be
	! mapped?
	!___________________________________________________________________
	fwind=(jblock-1)*maxtr_per_block

	!READ IN THE HEADERS IN THIS BLOCK
	call sep_get_val_headers("in", fwind+1, ntr_per_block, headers)
	!_______________________________________________________________________
	! echo out how many headers are out there on this file.
	!_______________________________________________________________________
	!___________________________________________________________________
	! grab a subset of desired headers using the header key words.
	!___________________________________________________________________
	gy    = headers (gy_key,    :)
	tracf = headers (tracf_key, :)
	nodeid= headers (node_key, :)

	!IF WE HAVE A DATA RECORD NUMBER WE NEED TO GRAB THE
	!CORRESPONDING TRACE NUMBERS. THE HEADER BLOCK IS
	!REAL, DRN ARE INTS. WE THEREFORE NEED TO CONVERT

	if(have_drn) then
		drn(1:ntr_per_block) = transfer (headers (drn_key, :ntr_per_block), 0, ntr_per_block)
	
		do i=1,ntr_per_block
			!___________________________________________________________________
			! now let's try to read ntr_per_block traces of seismic data.
			!___________________________________________________________________
			!SEEK TO THE BEGINING OF THE GIVEN TRACE
			if (sseek_block ("in", drn(i)-1, 4*nt, 0) < 0) call seperr("trouble seeking trace")
			if (sreed ("in", d(:,i),   4*nt) .ne. 4*nt)    call seperr("trouble reading trace ")
			drn(i)=i+fwind !RESET DRN TO ITS NEW POSITION
		end do
	else !TRACES ARE IN ORDER SO JUST READ THEN
	!___________________________________________________________________
	! now let's try to read ntr_per_block traces of seismic data.
	!___________________________________________________________________
		if (sreed("in", d, 4*nt*ntr_per_block) .ne. 4*nt*ntr_per_block) call seperr("trouble reading trace block")
	end if

	!___________________________________________________________________
	! flip the polarity of the trace if gy is less than gyflip
	!___________________________________________________________________
	do itrc = 1, ntr_per_block
		if (verb) write (stderr,*) 'node=', nodeid(itrc), 'gy=', gy (itrc), 'component=', tracf (itrc)
		if (gy (itrc) .lt. gyflip .and. ((tracf (itrc) .lt. VX+eps .and. tracf (itrc) .gt. VX-eps) .or. (tracf (itrc) .lt. VY+eps .and. tracf (itrc) .gt. VY-eps))) then
			if (verb) write (stderr,*) 'flipping polarity at node', nodeid(itrc), 'component=', tracf (itrc)
			d (:,itrc) = d (:,itrc) * -1.
		endif
	end do
	
	!LETS ACTUALLY WRITE OUT OUR CHANGES TO THE OUTPUT
	headers (gy_key,   1:ntr_per_block) = gy    (:ntr_per_block)
	headers (tracf_key,1:ntr_per_block) = tracf (:ntr_per_block)
	
	if (have_drn) headers (drn_key,1:ntr_per_block) = transfer (drn(:ntr_per_block), 0., ntr_per_block)
	
	!PUT THE HEADERS BACK OUT
	call sep_put_val_headers ("out", fwind+1, ntr_per_block, headers)
	
	!WRITE THE DATA
	if (srite ("out", d, 4*nt*ntr_per_block) .ne. 4*nt*ntr_per_block) call seperr("trouble writing out data")

end do loop_over_blocks

if (verb) write (stderr,*) 'fix_pol: normal completion'

end program fix_pol

