!rotate_vxvy
!______________________________________________________________________
! rotate horizontal components of a receiver gather after NMO with vel=water and sum along offset axis
!______________________________________________________________________
! USAGE:
! rotate_vxvy.x < in.H > out.H pars
!______________________________________________________________________
program rotate_vxvy
implicit none
!______________________________________________________________________
! allocatable arrays
!______________________________________________________________________
real,		allocatable, dimension(:,:) :: headers				! header block, expect all keys to be floats
real,		allocatable, dimension(:,:) :: data					! input seismic data
real,		allocatable, dimension(:,:) :: vx_data, vy_data		! X and Y data components
real,		allocatable, dimension(:)   :: vx_rot, vy_rot		! X and Y data components rotated at all angles
real,		allocatable, dimension(:)   :: nodeid				! receiver box number
real,		allocatable, dimension(:)   :: tracf				! component number
integer,	allocatable, dimension(:)   :: drn			! data record number
real,		allocatable, dimension(:)	:: tempvx
!______________________________________________________________________
! LOCATION OF VARIOUS KEYS IN THE HEADER STRUCTURES
!______________________________________________________________________
integer :: node_key, tracf_key, drn_key
!______________________________________________________________________
! simple variables
!______________________________________________________________________
integer :: nt	! number of time samples
real    :: ot
real    :: dt
integer :: nkeys			! number of keys
integer :: ntrace			! total number of input traces
integer :: maxsize			! memory available for storage in words.
integer :: maxtr_per_block	! maximum number of traces in a block of data.
integer :: ntr_per_block	! number of traces in current block of data.
integer :: nblocks			! number of blocks of needed to read all the traces.
integer :: jblock			! block index
integer :: fwind			! first trace of the data window to be read.
integer :: ierr				! return value checker
logical :: have_drn			! WHETHER OR NOT DRN EXISTS
real	:: node_num			! id of node we want to rotate
logical :: verb
real	:: angle
real	:: eps=1e-4
integer	:: ncomps			! number of data components
real, parameter :: VX=1, VY=2, deg2rad=4.*atan(1.)/180.
!______________________________________________________________________
! control variables
!______________________________________________________________________
integer :: allocation_status ! return code for Fortran90 allocate command
integer :: stderr=0 ! standard error unit
!______________________________________________________________________
! DUMMY VARIALBES
!______________________________________________________________________
real :: o,dd
integer :: itrc ! LOOP VARIABLE
character(len=1025) :: label

!EXTERNAL SHOULD BE USED FOR ALL C CALLS, SOME F90 COMPILERS GET MAD WITHOUT IT
integer,external :: getch				! SEPlib get param from parameter file
integer,external :: hetch				! SEPlib get param from history file
integer,external :: fetch				! SEPlib get from either parameter or history file
integer,external :: putch				! SEPlib put to history file
integer,external :: sep_get_number_keys	! SEPlib number of keys
integer,external :: sep_get_key_index	! SEPlib index of given header key
integer,external :: sseek_block			! SEPlib SEEK FILE ROUTINE
integer,external :: sreed				! SEPlib READ ROUTINE
integer,external :: srite				! SEPlib WRITE ROUTINE
integer,external :: sep_get_header_axis_par		! SEPlib HEADER AXIS ROUTINE
integer,external :: sep_get_number_header_axes	! SEPlib HEADER AXIS ROUTINE

!______________________________________________________________________
! self documentation
!______________________________________________________________________
call sep_add_doc_line("NAME")
call sep_add_doc_line("    rotate_vxvy - rotate the two horizontal components of a particular receiver around the vertical axis by a specified of angle.")
call sep_add_doc_line("")
call sep_add_doc_line("SYNOPSIS")
call sep_add_doc_line("    rotate_vxvy.x < in.H > out.H pars")
call sep_add_doc_line("")
call sep_add_doc_line("INPUT PARAMETERS")
call sep_add_doc_line("    verb - integer")
call sep_add_doc_line("               [0] Whether (1) or not(0) to be verbose")
call sep_add_doc_line("")
call sep_add_doc_line("    maxsize - integer")
call sep_add_doc_line("               [10] Number of megabytes of traces to hold")
call sep_add_doc_line("")
call sep_add_doc_line("    nodeid - integer")
call sep_add_doc_line("               [0] -  ocean-bottom node (receiver) physical box number to rotate")
call sep_add_doc_line("")
call sep_add_doc_line("    angle - real")
call sep_add_doc_line("               [0.]  angle of rotation to apply to horizontal geophone components")
call sep_add_doc_line("")
call sep_add_doc_line("    ncomps - integer")
call sep_add_doc_line("               [4]  number of data components")
call sep_add_doc_line("")
call sep_add_doc_line("DESCRIPTION")
call sep_add_doc_line("     Program assumes a sep3d input, where the horizontal components' field is named 'Tracf'.")
call sep_add_doc_line("     It assumes that the components are ordered so that the X horizontal geophone is one trace before the Y horizontal geophone")
call sep_add_doc_line("     The program rotates the horizontal components of the specified node (receiver) with the specified angle.")
call sep_add_doc_line("")

!_______________________________________________________________
! initialize the input seismic data file.
! do i need to specify any arguments?
!_______________________________________________________________
!TO INITIALIZE SEP IO/ PARA HANDLING
call initpar()
!SO SELF DOC WORKS
call doc(SOURCE)
!TO INITIALiZE SEP3D IO
call init_3d()
!_______________________________________________________________
! read in important parameters from the history or parameter file
!_______________________________________________________________
if (hetch ('n1',       'd', nt)       <= 0) call seperr('cant read n1 from history file!')
if (hetch ('o1',       'f', ot)       <= 0) call seperr('cant read o1 from history file!')
if (hetch ('d1',       'f', dt)       <= 0) call seperr('cant read d1 from history file!')
if (getch ('maxsize',   'd', maxsize )<= 0) maxsize=250
if (getch ('nodeid',   'f', node_num) <= 0) node_num=0. ! box number of node that we want to rotate
if (getch ('angle',     'f', angle)   <= 0) angle=0.
if (getch ('ncomps',     'd', ncomps)   <= 0) ncomps=4.
if (getch ('verb',     'd', verb)     <= 0) verb=.false.

if (verb) then
	write (stderr,*) 'nodeid=', node_num
	write (stderr,*) 'angle=', angle
endif

angle = angle * deg2rad

!THE NUMBER OF TRACES IS IN THE HEADER FILE
!IT DOESN'T HAVE TO BE THE SAME IS IN THE HISTORY FILE
!THEREFORE
if(sep_get_number_header_axes("in",ierr) .ne. 0) call seperr("trouble obtaining the number of header axes")

if(ierr .ne. 2) call seperr("this program expects a 2-D header file")

! read from axis 2: number of traces
if(sep_get_header_axis_par("in", 2, ntrace, o, dd, label) .ne. 0) call seperr("trouble obtaining the number of headers")

!______________________________________________________________________
! determine how many traces can fit into maxsize words of memory.
! determine how many blocks will need to be read in to read through
! all the input data.
!______________________________________________________________________
maxtr_per_block=((maxsize*1024.*1024.-1.)/(ntrace*nt)+1)*ncomps
nblocks=(ntrace-1)/maxtr_per_block+1

if (verb) write (stderr,*) 'node=', node_num, 'angle=', angle, 'maxsize=', maxsize, 'ntrace=', ntrace, 'nt=', nt, 'maxtr_per_block=', maxtr_per_block

!_______________________________________________________________
! read in important parameters from the hff file
!_______________________________________________________________
if (sep_get_number_keys("in",nkeys) .ne. 0) call seperr("trouble obtaining number of keys from hff")

if (sep_get_key_index("in", "Nodeid", node_key) .ne. 0) call seperr("trouble obtaining Nodeid key number")
if (sep_get_key_index("in", "Tracf", tracf_key) .ne. 0) call seperr("trouble obtaining Tracf key number")

ierr = sep_get_key_index("in","data_record_number",drn_key)

if (ierr .ne. 0) then !DRN DOESN'T EXIST;

	if (getch ("same_record_number","d",ierr) == 1) then !IF SAME RECORD NUMBER IN HISTORY
		if(ierr==0) call seperr("no data_record_number but traces out of order with data")
	end if
	have_drn=.false.
else
	have_drn=.true.
end if

!______________________________________________________________________
! ALLOCATE arrays
! note that output arrays are of length nt (the same as the input).
! arrays used in Fourier transform are of length nt_fft >= nt.
!______________________________________________________________________
allocate (data (0:nt-1,maxtr_per_block), &
	tempvx (0:nt-1), &
	headers (nkeys,maxtr_per_block), &
	nodeid (maxtr_per_block), &
	tracf (maxtr_per_block), &
	drn (maxtr_per_block), &
	STAT=allocation_status)

if(allocation_status /= 0) THEN
	write (stderr,*) 'space for data,headers,nodeid,tracf could not be allocated!'
	write (stderr,*) 'program aborted!'
	stop 9666
endif

!_______________________________________________________________________
! write out the history files defining the size of the output.
!_______________________________________________________________________
!LETS DEAL WITH THE OUTPUT
!WE ARE GOING TO REORDER THE TRACES, SO LETS TELL THE PROGRAM THAT
!THE TRACES ARE IN THE SAME ORDER

if(0 .ne. putch ("same_record_number","d",1)) call seperr("trouble putching same_record_number")

!IN THIS PROGRAM THE OUTPUT IS THE SAME SIZE OF
!THE INPUT. THE HISTORY FILE BY DEFAULT IS THE SAME
!SIZE AS THE INPUT SO WE DON'T NEED TO DO ANY MANIPULATION
!
!THE HFF FILE DOES NEED TO BE CREATED. SAYING WE
!HAVE THE SAME KEYS. WE CAN JUST COPY THE HFF FILE
call sep_copy_hff("in","out")

!_____________________________________________________________
! close sep data files
!_____________________________________________________________
call hclose()

!___________________________________________________________________
! loop over the blocks of data.
!___________________________________________________________________
loop_over_blocks: do jblock=1,nblocks

	!___________________________________________________________________
	! calculate the number of traces in this block.
	!___________________________________________________________________
	ntr_per_block=maxtr_per_block
	if (jblock == nblocks) ntr_per_block = ntrace-(nblocks-1)*maxtr_per_block
	!___________________________________________________________________
	! grab the next ntr_per_block trace headers from the trace header file.
	! 'in' = standard in
	! headers = a fortran90 structure
	! fwind = the sequential trace position in the trace header file.
	!___________________________________________________________________
	fwind=(jblock-1)*maxtr_per_block

	!READ IN THE HEADERS IN THIS BLOCK
	call sep_get_val_headers("in", fwind+1, ntr_per_block, headers)
	!_______________________________________________________________________
	! echo out how many headers are out there on this file.
	!_______________________________________________________________________
	!___________________________________________________________________
	! grab a subset of desired headers using the header key words.
	!___________________________________________________________________
	nodeid	= headers (node_key,:)
	tracf	= headers (tracf_key,:)

	!IF WE HAVE A DATA RECORD NUMBER WE NEED TO GRAB THE
	!CORRESPONDING TRACE NUMBERS. THE HEADER BLOCK IS
	!REAL, DRN ARE INTS. WE THEREFORE NEED TO CONVERT

	if(have_drn) then
		drn(1:ntr_per_block) = transfer (headers (drn_key, :ntr_per_block), 0, ntr_per_block)
	
		do itrc=1,ntr_per_block
			!___________________________________________________________________
			! now let's try to read ntr_per_block traces of seismic data.
			!___________________________________________________________________
			!SEEK TO THE BEGINING OF THE GIVEN TRACE
			if (sseek_block ("in", drn(itrc)-1, 4*nt, 0) < 0)  call seperr("trouble seeking trace")
			if (sreed ("in", data (:,itrc),   4*nt) .ne. 4*nt) call seperr("trouble reading trace ")
			drn(itrc)=itrc+fwind !RESET DRN TO ITS NEW POSITION
		end do
	else !TRACES ARE IN ORDER SO JUST READ THEN
	!___________________________________________________________________
	! now let's try to read ntr_per_block traces of seismic data.
	!___________________________________________________________________
		if (sreed("in", data, 4*nt*ntr_per_block) .ne. 4*nt*ntr_per_block) call seperr("trouble reading trace block")
	end if

	!___________________________________________________________________
	! if this node is the one specified on the command line, 
	! Rotate the Vx and Vy components by the specified angle.
	!___________________________________________________________________

	do itrc = 1, ntr_per_block
		if (nodeid (itrc) .lt. node_num+eps .and. nodeid (itrc) .gt. node_num-eps) then

			if (tracf (itrc) .lt. VX+eps .and. tracf (itrc) .gt. VX-eps) then
				!if (verb) write (stderr,*) 'itrc=', itrc, 'node=', nodeid (itrc), 'tracf=', tracf (itrc)

				tempvx (:) = data (:, itrc)
				data (:, itrc)   = data (:, itrc) * cos (angle) - data (:, itrc+1) * sin (angle)
				data (:, itrc+1) = tempvx (:)     * sin (angle) + data (:, itrc+1) * cos (angle)
			endif
			
		endif
	end do
	
	!WRITE THE DATA
	if (srite ("out", data, 4*nt*ntr_per_block) .ne. 4*nt*ntr_per_block) call seperr("trouble writing out data")

	!PUT THE HEADERS BACK OUT
	call sep_put_val_headers ("out", fwind+1, ntr_per_block, headers)

end do loop_over_blocks


if (verb) write (stderr,*) 'rotate_vxvy: normal completion.'

end program rotate_vxvy

